
//Chris' WinGlide enhancements
//Copr. 1998, Chris Dohnal (cwdohnal@ucdavis.edu)

#include <windows.h>
#include <glide.h>
#include "inifile.h"
#include "dib.h"
#include "msgbox.h"
#include "gamma.h"
#include "pcopy.h"
#include "pointers.h"
#include "main.h"
#include "sysmenu.h"

BOOL B2WGInitialize(HWND, DWORD, DWORD);
VOID B2WGShutdown(VOID);
VOID B2WGCopy3DfxFrameBuffer(VOID);

static HWND ghWnd;
static WNDPROC gpPrevWndProc = NULL;
static BOOL gbWindowWasDestroyed = FALSE;
static LRESULT CALLBACK SubclassWndProc(HWND, UINT, WPARAM, LPARAM);


BOOL B2WGInitialize(HWND hWnd, DWORD dwWidth, DWORD dwHeight) {	
	BOOL bRet;

	//Set up the global hWnd
	ghWnd = hWnd;

	//Create the DIB
	//B2 WinGlide must use the size of the window on the 3Dfx card for memory allocation so that resizing works
	bRet = CreateDIB(hWnd, gdwGlideScreenWidth, gdwGlideScreenHeight);
	if (bRet == FALSE) {
		ErrorLoadingWinGlideMessageBox(NULL, "Could not create DIB section.");
	}

	//Allocate memory for the gamma correction table
	bRet = AllocateGammaCorrectionTable();
	if (bRet == FALSE) {
		ErrorLoadingWinGlideMessageBox(NULL, "Could not create gamma correction table.");

		return FALSE;
	}

	//Calculate the gamma correction table
	CalculateGammaCorrectionTable(gRedGamma, gGreenGamma, gBlueGamma);
	
	//See if the subclassing used to provide window resizing support should be disabled
	if (gbDisableResizeSubclass == FALSE) {
		//Reset the window was destroyed flag
		gbWindowWasDestroyed = FALSE;
		//Subclass the window
		gpPrevWndProc = (WNDPROC)GetWindowLong(hWnd, GWL_WNDPROC);
		SetWindowLong(hWnd, GWL_WNDPROC, (LONG)SubclassWndProc);
	}
	else {
		//Set this to NULL just in case
		gpPrevWndProc = NULL;
	}

	//Do not add commands to the system menu is subclassing is disabled
	if (gbDisableResizeSubclass == FALSE) {
		//If system menu options are enabled, add options to the system menu
		if (gbSystemMenuOptions == TRUE) {
			//Add WinGlide options to the system menu
			AddWinGlideSystemMenuOptions(hWnd, WINGLIDE_MODE_B2);
		}
	}

	return TRUE;
}

VOID B2WGShutdown(VOID) {
	//See if the window was destroyed, making its handle invalid
	if (gbWindowWasDestroyed == TRUE) {
		//Don't try to unsubclass if the window has been destroyed
		gpPrevWndProc = NULL;
	}
	//Do shutdown and cleanup stuff if the window was not destroyed
	if (gpPrevWndProc != NULL) {
		//If system menu options are enabled, remove options from the system menu
		if (gbSystemMenuOptions == TRUE) {
			//Remove WinGlide options from the system menu
			RemoveWinGlideSystemMenuOptions(ghWnd);
		}
	}

	//Unsubclass the window if it was subclassed
	if (gpPrevWndProc != NULL) {
		SetWindowLong(ghWnd, GWL_WNDPROC, (LONG)gpPrevWndProc);
		gpPrevWndProc = NULL;
	}

	//Delete the DIB
	FreeDIB();

	//Free the gamma correction table
	FreeGammaCorrectionTable();
	
	return;
}

VOID B2WGCopy3DfxFrameBuffer(VOID) {
	FxBool fxbRet;
	GrLfbInfo_t grLFBInfo;
	BYTE *p3DfxLFBStart;
	BYTE *pDIB;
	LONG lDIBPitch;
	DWORD *pGammaTable;
	LONG lAlignedWidth;

	//Check the range of these values and make sure they are valid
	if (gdwScreenWidth > gdwGlideScreenWidth) { gdwScreenWidth = gdwGlideScreenWidth; }
	if (gdwScreenHeight > gdwGlideScreenHeight) { gdwScreenHeight = gdwGlideScreenHeight; }
	if (gdwScreenWidth < 0) { gdwScreenWidth = 0; }
	if (gdwScreenHeight < 0) { gdwScreenHeight = 0; }

	//Get information about the DIB
	pDIB = GetDIBBuffer();
	if (pDIB == NULL) {
		return;
	}
	lDIBPitch = GetDIBPitch();

	//Get the pointer to the gamma correction table
	if (gbEnableGammaCorrection == TRUE) {
		pGammaTable = GetGammaCorrectionTablePointer();
		if (pGammaTable == NULL) {
			return;
		}
	}

	//Lock the 3Dfx LFB
	ZeroMemory(&grLFBInfo, sizeof(grLFBInfo));
	grLFBInfo.size = sizeof(grLFBInfo);
	fxbRet = _grLfbLock(GR_LFB_READ_ONLY, GR_BUFFER_BACKBUFFER, 0,
		GR_ORIGIN_ANY, FXFALSE, &grLFBInfo);
	if (fxbRet == FXFALSE) {
		//Could not lock the 3Dfx LFB
		return;
	}

	//Setup the pointer to the 3Dfx LFB
	if (gOriginLocation == GR_ORIGIN_LOWER_LEFT) {
		p3DfxLFBStart = (BYTE *)grLFBInfo.lfbPtr +
			(grLFBInfo.strideInBytes * (gdwScreenHeight - 1));
	}
	else {
		p3DfxLFBStart = (BYTE *)grLFBInfo.lfbPtr +
			(grLFBInfo.strideInBytes * (gdwGlideScreenHeight - 1));
	}
	
	//Get the aligned width to copy
	lAlignedWidth = ((gdwScreenWidth + 3) & ~3) >> 2;
	
	//See if gamma correction is enabled
	if (gbEnableGammaCorrection == TRUE) {
		//Copy the frame from the 3Dfx to the DIB
		if (gbUseMMX == TRUE) {
			MMXGammaPitchCopy8(p3DfxLFBStart, pDIB,
				-(LONG)grLFBInfo.strideInBytes, lDIBPitch,
				lAlignedWidth, gdwScreenHeight,
				pGammaTable);
		}
		else {
			GammaPitchCopy8(p3DfxLFBStart, pDIB,
				-(LONG)grLFBInfo.strideInBytes, lDIBPitch,
				lAlignedWidth, gdwScreenHeight,
				pGammaTable);
		}

	}
	else {
		//Copy the frame from the 3Dfx to the DIB
		if (gbUseMMX == TRUE) {
			MMXPitchCopy8(p3DfxLFBStart, pDIB,
				-(LONG)grLFBInfo.strideInBytes, lDIBPitch,
				lAlignedWidth, gdwScreenHeight);
		}
		else {
			FPUPitchCopy8(p3DfxLFBStart, pDIB,
				-(LONG)grLFBInfo.strideInBytes, lDIBPitch,
				lAlignedWidth, gdwScreenHeight);
		}
	}	

	//Unlock the 3Dfx LFB
	fxbRet = _grLfbUnlock(GR_LFB_READ_ONLY, GR_BUFFER_BACKBUFFER);
	if (fxbRet == FXFALSE) {
		//Could not unlock the 3Dfx LFB
		return;
	}

	//Put the DIB on the screen
	DrawDIB();

	return;
}

static LRESULT CALLBACK SubclassWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	RECT rect;

	switch (message) {
	case WM_SYSCOMMAND:
		//See if system menu options are enabled
		if (gbSystemMenuOptions == TRUE) {
			BOOL bRet;

			//Handle WinGlide system menu options
			bRet = HandleWinGlideSystemMenuOptions(hWnd, wParam, lParam);

			//Break if the message was handled
			if (bRet == TRUE) {
				break;
			}
			//If the message was not handled, pass it on to the previous window procedure
		}

		//Send the WM_SYSCOMMAND message to the previous window procedure
		return CallWindowProc(gpPrevWndProc, hWnd, message, wParam, lParam);

		break;

	case WM_MOVE:
		//Send the message to the previous window procedure
		return CallWindowProc(gpPrevWndProc, hWnd, message, wParam, lParam);

		break;

	case WM_SIZE:
		//Send the message to the previous window procedure
		CallWindowProc(gpPrevWndProc, hWnd, message, wParam, lParam);

		GetClientRect(ghWnd, &rect);
		gdwScreenWidth = rect.right;
		gdwScreenHeight = rect.bottom;
		if (gdwScreenWidth > gdwGlideScreenWidth) { gdwScreenWidth = gdwGlideScreenWidth; }
		if (gdwScreenHeight > gdwGlideScreenHeight) { gdwScreenHeight = gdwGlideScreenHeight; }
		if (gdwScreenWidth < 0) { gdwScreenWidth = 0; }
		if (gdwScreenHeight < 0) { gdwScreenHeight = 0; }

		break;

	case WM_DESTROY:
		//Set the flag indicating that the window was destroyed
		gbWindowWasDestroyed = TRUE;

		//Send the message to the previous window procedure
		return CallWindowProc(gpPrevWndProc, hWnd, message, wParam, lParam);

		break;

	default:
		return CallWindowProc(gpPrevWndProc, hWnd, message, wParam, lParam);
	}
	return 0;	
}
