
//Chris' WinGlide enhancements
//Copr. 1998, Chris Dohnal (cwdohnal@ucdavis.edu)

#include <windows.h>
#include <glide.h>
#include "inifile.h"
#include "dib.h"
#include "msgbox.h"
#include "gamma.h"
#include "pcopy.h"
#include "pointers.h"
#include "main.h"
#include "sysmenu.h"

BOOL StretchBltWGInitialize(HWND, DWORD, DWORD);
VOID StretchBltWGInitializeWGShutdown(VOID);
VOID StretchBltWGInitializeWGCopy3DfxFrameBuffer(VOID);

static HWND ghWnd;
static WNDPROC gpPrevWndProc = NULL;
static LONG glOldWindowStyle;
static UINT guiPrevCloseMenuState;
static BOOL gbWindowWasDestroyed = FALSE;
static LRESULT CALLBACK SubclassWndProc(HWND, UINT, WPARAM, LPARAM);


BOOL StretchBltWGInitialize(HWND hWnd, DWORD dwWidth, DWORD dwHeight) {	
	BOOL bRet;
	RECT rect;
	int iExtraWidth, iExtraHeight;

	//Set up the global hWnd
	ghWnd = hWnd;

	//Create the DIB
	bRet = CreateDIB(hWnd, dwWidth, dwHeight);
	if (bRet == FALSE) {
		ErrorLoadingWinGlideMessageBox(NULL, "Could not create DIB section.");
	}

	//Allocate memory for the gamma correction table
	bRet = AllocateGammaCorrectionTable();
	if (bRet == FALSE) {
		ErrorLoadingWinGlideMessageBox(NULL, "Could not create gamma correction table.");

		return FALSE;
	}

	//Calculate the gamma correction table
	CalculateGammaCorrectionTable(gRedGamma, gGreenGamma, gBlueGamma);
	
	//Reset the window was destroyed flag
	gbWindowWasDestroyed = FALSE;
	//Subclass the window
	gpPrevWndProc = (WNDPROC)GetWindowLong(hWnd, GWL_WNDPROC);
	SetWindowLong(hWnd, GWL_WNDPROC, (LONG)SubclassWndProc);

	//Hide the window
	ShowWindow(hWnd, SW_HIDE);
	
	//Change the window style
	glOldWindowStyle = GetWindowLong(hWnd, GWL_STYLE);
	SetWindowLong(hWnd, GWL_STYLE, WS_OVERLAPPEDWINDOW);

	//If system menu options are enabled, add options to the system menu
	if (gbSystemMenuOptions == TRUE) {
		//Add WinGlide options to the system menu
		AddWinGlideSystemMenuOptions(hWnd, WINGLIDE_MODE_STRETCH_BLT);
	}

	//Disable the close command on the system menu if the previous window style did not have a
	//system menu
	if (!(glOldWindowStyle & WS_SYSMENU)) {
		HMENU hMenu;

		hMenu = GetSystemMenu(hWnd, FALSE);
		guiPrevCloseMenuState = GetMenuState(hMenu, SC_CLOSE, 0);
		EnableMenuItem(hMenu, SC_CLOSE, MF_GRAYED);
	}
	
	//Resize the window so the client area is lWidth by lHeight
	GetWindowRect(hWnd, &rect);
	iExtraWidth = (GetSystemMetrics(SM_CXFRAME) * 2);
	iExtraHeight = (GetSystemMetrics(SM_CXFRAME) * 2) + GetSystemMetrics(SM_CYCAPTION);
	MoveWindow(hWnd, rect.left, rect.top, dwWidth + iExtraWidth, dwHeight + iExtraHeight, TRUE);
	
	//Show the window
	ShowWindow(hWnd, SW_SHOW);
	
	//Update the window
	InvalidateRect(hWnd, NULL, FALSE);
	UpdateWindow(hWnd);

	return TRUE;
}

VOID StretchBltWGShutdown(VOID) {
	//See if the window was destroyed, making its handle invalid
	if (gbWindowWasDestroyed == TRUE) {
		//Don't try to unsubclass if the window has been destroyed
		gpPrevWndProc = NULL;
	}
	//Do shutdown and cleanup stuff if the window was not destroyed
	if (gpPrevWndProc != NULL) {
		//If system menu options are enabled, remove options from the system menu
		if (gbSystemMenuOptions == TRUE) {
			//Remove WinGlide options from the system menu
			RemoveWinGlideSystemMenuOptions(ghWnd);
		}
	}

	//Unsubclass the window if it was subclassed
	if (gpPrevWndProc != NULL) {
		SetWindowLong(ghWnd, GWL_WNDPROC, (LONG)gpPrevWndProc);
		gpPrevWndProc = NULL;
	}

	//Delete the DIB
	FreeDIB();

	//Free the gamma correction table
	FreeGammaCorrectionTable();
	
	return;
}

VOID StretchBltWGCopy3DfxFrameBuffer(VOID) {
	FxBool fxbRet;
	GrLfbInfo_t grLFBInfo;
	BYTE *p3DfxLFBStart;
	BYTE *pDIB;
	LONG lDIBPitch;
	DWORD *pGammaTable;
	LONG lAlignedWidth;

	//Check the range of these values and make sure they are valid
	if (gdwScreenWidth > gdwGlideScreenWidth) { gdwScreenWidth = gdwGlideScreenWidth; }
	if (gdwScreenHeight > gdwGlideScreenHeight) { gdwScreenHeight = gdwGlideScreenHeight; }
	if (gdwScreenWidth < 0) { gdwScreenWidth = 0; }
	if (gdwScreenHeight < 0) { gdwScreenHeight = 0; }

	//Get information about the DIB
	pDIB = GetDIBBuffer();
	if (pDIB == NULL) {
		return;
	}
	lDIBPitch = GetDIBPitch();

	//Get the pointer to the gamma correction table
	if (gbEnableGammaCorrection == TRUE) {
		pGammaTable = GetGammaCorrectionTablePointer();
		if (pGammaTable == NULL) {
			return;
		}
	}

	//Lock the 3Dfx LFB
	ZeroMemory(&grLFBInfo, sizeof(grLFBInfo));
	grLFBInfo.size = sizeof(grLFBInfo);
	fxbRet = _grLfbLock(GR_LFB_READ_ONLY, GR_BUFFER_BACKBUFFER, 0,
		GR_ORIGIN_ANY, FXFALSE, &grLFBInfo);
	if (fxbRet == FXFALSE) {
		//Could not lock the 3Dfx LFB
		return;
	}

	//Setup the pointer to the 3Dfx LFB
	if (gOriginLocation == GR_ORIGIN_LOWER_LEFT) {
		p3DfxLFBStart = (BYTE *)grLFBInfo.lfbPtr +
			(grLFBInfo.strideInBytes * (gdwScreenHeight - 1));
	}
	else {
		p3DfxLFBStart = (BYTE *)grLFBInfo.lfbPtr +
			(grLFBInfo.strideInBytes * (gdwGlideScreenHeight - 1));
	}
	
	//Get the aligned width to copy
	lAlignedWidth = ((gdwScreenWidth + 3) & ~3) >> 2;
	
	//See if gamma correction is enabled
	if (gbEnableGammaCorrection == TRUE) {
		//Copy the frame from the 3Dfx to the DIB
		if (gbUseMMX == TRUE) {
			MMXGammaPitchCopy8(p3DfxLFBStart, pDIB,
				-(LONG)grLFBInfo.strideInBytes, lDIBPitch,
				lAlignedWidth, gdwScreenHeight,
				pGammaTable);
		}
		else {
			GammaPitchCopy8(p3DfxLFBStart, pDIB,
				-(LONG)grLFBInfo.strideInBytes, lDIBPitch,
				lAlignedWidth, gdwScreenHeight,
				pGammaTable);
		}

	}
	else {
		//Copy the frame from the 3Dfx to the DIB
		if (gbUseMMX == TRUE) {
			MMXPitchCopy8(p3DfxLFBStart, pDIB,
				-(LONG)grLFBInfo.strideInBytes, lDIBPitch,
				lAlignedWidth, gdwScreenHeight);
		}
		else {
			FPUPitchCopy8(p3DfxLFBStart, pDIB,
				-(LONG)grLFBInfo.strideInBytes, lDIBPitch,
				lAlignedWidth, gdwScreenHeight);
		}
	}	

	//Unlock the 3Dfx LFB
	fxbRet = _grLfbUnlock(GR_LFB_READ_ONLY, GR_BUFFER_BACKBUFFER);
	if (fxbRet == FXFALSE) {
		//Could not unlock the 3Dfx LFB
		return;
	}

	//Put the DIB on the screen
	DrawStretchedDIB();

	return;
}

static LRESULT CALLBACK SubclassWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	switch (message) {
	case WM_SYSCOMMAND:
		//See if system menu options are enabled
		if (gbSystemMenuOptions == TRUE) {
			BOOL bRet;

			//Handle WinGlide system menu options
			bRet = HandleWinGlideSystemMenuOptions(hWnd, wParam, lParam);

			//Break if the message was handled
			if (bRet == TRUE) {
				break;
			}
			//If the message was not handled, pass it on to the previous window procedure
		}

		//Send the WM_SYSCOMMAND message to the previous window procedure
		return CallWindowProc(gpPrevWndProc, hWnd, message, wParam, lParam);

		break;

	case WM_MOVE:
		//Send the message to the previous window procedure
		return CallWindowProc(gpPrevWndProc, hWnd, message, wParam, lParam);

		break;

	case WM_SIZE:
		//Send the message to the previous window procedure
		CallWindowProc(gpPrevWndProc, hWnd, message, wParam, lParam);

		//Put the DIB on the screen
		DrawStretchedDIB();

		break;

	case WM_DESTROY:
		//Set the flag indicating that the window was destroyed
		gbWindowWasDestroyed = TRUE;

		//Send the message to the previous window procedure
		return CallWindowProc(gpPrevWndProc, hWnd, message, wParam, lParam);

		break;

	default:
		return CallWindowProc(gpPrevWndProc, hWnd, message, wParam, lParam);
	}
	return 0;	
}

